# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

try:
    # pybind11 only supports cPickle protocol 2+ (-1 in pickle.dumps)
    # (automatically satisfied in Python 3; this is just to keep testing Python 2.7)
    import cPickle as pickle
except ImportError:
    import pickle

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2  # noqa: F401

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_forms():
    form = ak._v2.forms.NumpyForm("float64")
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert form.inner_shape == ()
    assert form.itemsize == 8
    assert form.primitive == "float64"
    assert form.has_identifier is False
    assert form.parameters == {}
    assert form.form_key is None
    assert json.loads(form.to_json()) == {
        "class": "NumpyArray",
        "inner_shape": [],
        "primitive": "float64",
        "has_identifier": False,
        "parameters": {},
        "form_key": None,
    }
    assert json.loads(str(form)) == {
        "class": "NumpyArray",
        "primitive": "float64",
    }

    form = ak._v2.forms.NumpyForm(
        "float64",
        [1, 2, 3],
        has_identifier=True,
        parameters={"hey": ["you", {"there": 3}]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert form.inner_shape == (1, 2, 3)
    assert form.itemsize == 8
    assert form.primitive == "float64"
    assert form.has_identifier is True
    assert form.parameters == {"hey": ["you", {"there": 3}]}
    assert form.form_key == "yowzers"
    assert json.loads(form.to_json()) == {
        "class": "NumpyArray",
        "inner_shape": [1, 2, 3],
        "primitive": "float64",
        "has_identifier": True,
        "parameters": {"hey": ["you", {"there": 3}]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "NumpyArray",
        "inner_shape": [1, 2, 3],
        "primitive": "float64",
        "has_identifier": True,
        "parameters": {"hey": ["you", {"there": 3}]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.BitMaskedForm(
        "i8",
        ak._v2.forms.NumpyForm("float64"),
        True,
        False,
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "BitMaskedArray",
        "mask": "i8",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "valid_when": True,
        "lsb_order": False,
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "BitMaskedArray",
        "mask": "i8",
        "content": "float64",
        "valid_when": True,
        "lsb_order": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.EmptyForm(
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "EmptyArray",
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "EmptyArray",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.IndexedForm(
        "i64",
        ak._v2.forms.NumpyForm("float64"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "IndexedArray",
        "index": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "IndexedArray",
        "index": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.IndexedOptionForm(
        "i64",
        ak._v2.forms.NumpyForm("float64"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "IndexedOptionArray",
        "index": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "IndexedOptionArray",
        "index": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.ListForm(
        "i64",
        "i64",
        ak._v2.forms.NumpyForm("float64"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "ListArray",
        "starts": "i64",
        "stops": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "ListArray",
        "starts": "i64",
        "stops": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.ListOffsetForm(
        "i64",
        ak._v2.forms.NumpyForm("float64"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "ListOffsetArray",
        "offsets": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "ListOffsetArray",
        "offsets": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.RecordForm(
        [ak._v2.forms.NumpyForm("float64"), ak._v2.forms.NumpyForm("bool")],
        ["one", "two"],
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "RecordArray",
        "contents": {
            "one": {
                "class": "NumpyArray",
                "inner_shape": [],
                "primitive": "float64",
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
            "two": {
                "class": "NumpyArray",
                "inner_shape": [],
                "primitive": "bool",
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
        },
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "RecordArray",
        "contents": {
            "one": "float64",
            "two": "bool",
        },
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.RecordForm(
        [ak._v2.forms.NumpyForm("float64"), ak._v2.forms.NumpyForm("bool")],
        None,
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "RecordArray",
        "contents": [
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "primitive": "float64",
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "primitive": "bool",
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
        ],
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "RecordArray",
        "contents": ["float64", "bool"],
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.RegularForm(
        ak._v2.forms.NumpyForm("float64"),
        10,
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "RegularArray",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "size": 10,
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "RegularArray",
        "content": "float64",
        "size": 10,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.UnionForm(
        "i8",
        "i64",
        [ak._v2.forms.NumpyForm("float64"), ak._v2.forms.NumpyForm("bool")],
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i64",
        "contents": [
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "primitive": "float64",
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "primitive": "bool",
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
        ],
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i64",
        "contents": ["float64", "bool"],
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.UnmaskedForm(
        ak._v2.forms.NumpyForm("float64"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "UnmaskedArray",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "UnmaskedArray",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }

    form = ak._v2.forms.VirtualForm(
        ak._v2.forms.NumpyForm("float64"),
        True,
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak._v2.forms.from_json(form.to_json()) == form
    assert json.loads(form.to_json()) == {
        "class": "VirtualArray",
        "form": {
            "class": "NumpyArray",
            "inner_shape": [],
            "primitive": "float64",
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        },
        "has_length": True,
        "has_identifier": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "VirtualArray",
        "form": "float64",
        "has_length": True,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }


def fcn():
    return ak._v2.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))


def test_basic():
    generator = ak._v2.contents.FunctionGenerator(
        5, ak._v2.forms.NumpyForm("float64"), fcn
    )
    cache = {}

    virtualarray = ak._v2.contents.VirtualArray(generator, cache)
    assert virtualarray.peek_array is None
    assert virtualarray.array is not None
    assert ak.to_list(virtualarray.peek_array) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert ak.to_list(virtualarray.array) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert ak.to_list(cache[virtualarray.cache_key]) == [1.1, 2.2, 3.3, 4.4, 5.5]

    cache = None

    virtualarray = ak._v2.contents.VirtualArray(generator, cache)
    assert virtualarray.peek_array is None
    assert virtualarray.array is not None
    assert virtualarray.peek_array is None
    assert ak.to_list(virtualarray.array) == [1.1, 2.2, 3.3, 4.4, 5.5]


def test_slice():
    generator = ak._v2.contents.FunctionGenerator(
        3,
        None,
        lambda: v1_to_v2(
            ak.Array(
                [[1.1, 2.2, 3.3, 4.4, 5.5], [6.6, 7.7, 8.8], [100, 200, 300, 400]]
            ).layout
        ),
    )
    virtualarray = ak._v2.contents.VirtualArray(generator)

    assert isinstance(virtualarray, ak._v2.contents.VirtualArray)

    sliced = virtualarray[:-1]

    assert isinstance(sliced, ak._v2.contents.VirtualArray)

    assert isinstance(sliced[1], ak._v2.contents.NumpyArray)


def test_field():
    generator = ak._v2.contents.FunctionGenerator(
        None,
        None,
        lambda: v1_to_v2(
            ak.Array(
                [
                    {"x": 0.0, "y": []},
                    {"x": 1.1, "y": [1]},
                    {"x": 2.2, "y": [2, 2]},
                    {"x": 3.3, "y": [3, 3, 3]},
                ]
            ).layout
        ),
    )
    virtualarray = ak._v2.contents.VirtualArray(generator)

    assert isinstance(virtualarray, ak._v2.contents.VirtualArray)

    sliced = virtualarray["y"]
    assert isinstance(sliced, ak._v2.contents.VirtualArray)

    assert isinstance(sliced[1], ak._v2.contents.NumpyArray)


def test_fields():
    generator = ak._v2.contents.FunctionGenerator(
        None,
        None,
        lambda: v1_to_v2(
            ak.Array(
                [
                    {"x": 0.0, "y": []},
                    {"x": 1.1, "y": [1]},
                    {"x": 2.2, "y": [2, 2]},
                    {"x": 3.3, "y": [3, 3, 3]},
                ]
            ).layout
        ),
    )
    virtualarray = ak._v2.contents.VirtualArray(generator)

    assert isinstance(virtualarray, ak._v2.contents.VirtualArray)

    sliced = virtualarray[["y", "x"]]
    assert isinstance(sliced, ak._v2.contents.VirtualArray)

    assert isinstance(sliced[1], ak._v2.record.Record)


def test_single_level():
    template = v1_to_v2(
        ak.Array(
            [
                [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}],
                [],
                [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}],
            ]
        ).layout
    )
    generator = ak._v2.contents.FunctionGenerator(3, template.form, lambda: template)
    cache = {}
    virtualarray = ak._v2.contents.VirtualArray(generator, cache)

    a = virtualarray[2]
    assert isinstance(a, ak._v2.contents.RecordArray)
    assert len(cache) == 1
    assert ak.to_list(a) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    cache.clear()

    a = virtualarray[1:]
    assert isinstance(a, ak._v2.contents.VirtualArray)
    assert len(cache) == 0
    b = a[1]
    assert isinstance(b, ak._v2.contents.RecordArray)
    assert len(cache) >= 1
    assert ak.to_list(b) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    cache.clear()

    a = virtualarray[[0, 2, 1, 0]]
    assert isinstance(a, ak._v2.contents.VirtualArray)
    assert len(cache) == 0
    b = a[1]
    assert isinstance(b, ak._v2.contents.RecordArray)
    assert len(cache) >= 1
    assert ak.to_list(b) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    cache.clear()

    a = virtualarray[[False, True, True]]
    assert isinstance(a, ak._v2.contents.VirtualArray)
    assert len(cache) == 0
    b = a[1]
    assert isinstance(b, ak._v2.contents.RecordArray)
    assert len(cache) >= 1
    assert ak.to_list(b) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    cache.clear()

    a = virtualarray["x"]
    assert isinstance(a, ak._v2.contents.VirtualArray)
    assert len(cache) == 0
    b = a[2]
    assert isinstance(b, ak._v2.contents.NumpyArray)
    assert len(cache) >= 1
    assert ak.to_list(b) == [3.3, 4.4]
    cache.clear()

    a = virtualarray["y"]
    assert isinstance(a, ak._v2.contents.VirtualArray)
    assert len(cache) == 0
    b = a[2]
    assert isinstance(b, (ak._v2.contents.ListArray, ak._v2.contents.ListOffsetArray))
    assert len(cache) >= 1
    assert ak.to_list(b) == [[3, 3, 3], [4, 4, 4, 4]]
    cache.clear()

    a = virtualarray[:1, 1]
    assert isinstance(a, (ak._v2.contents.RecordArray, ak._v2.contents.IndexedArray))
    assert len(cache) >= 1
    assert ak.to_list(a) == [{"x": 1.1, "y": [1]}]
    cache.clear()

    a = virtualarray[::2, 1]
    assert isinstance(a, (ak._v2.contents.RecordArray, ak._v2.contents.IndexedArray))
    assert len(cache) >= 1
    assert ak.to_list(a) == [{"x": 1.1, "y": [1]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    cache.clear()


def test_iter():
    generator = ak._v2.contents.FunctionGenerator(
        5, ak._v2.forms.NumpyForm("float64"), fcn
    )
    cache = {}
    virtualarray = ak._v2.contents.VirtualArray(generator, cache)

    assert len(cache) == 0
    it = iter(virtualarray)
    assert len(cache) == 0
    assert next(it) == 1.1
    assert len(cache) == 1
    assert list(it) == [2.2, 3.3, 4.4, 5.5]
    assert len(cache) == 1


def test_nested_virtualness():
    counter = [0, 0]

    content = ak._v2.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )

    def materialize1():
        counter[1] += 1
        return content

    generator1 = ak._v2.contents.FunctionGenerator(
        len(content), content.form, materialize1
    )
    virtual1 = ak._v2.contents.VirtualArray(generator1)

    offsets = ak._v2.index.Index64(np.array([0, 3, 3, 5, 6, 10], dtype=np.int64))
    listarray = ak._v2.contents.ListOffsetArray(offsets, virtual1)

    def materialize2():
        counter[0] += 1
        return listarray

    generator2 = ak._v2.contents.FunctionGenerator(
        len(listarray), listarray.form, materialize2
    )
    virtual2 = ak._v2.contents.VirtualArray(generator2)

    assert counter == [0, 0]

    tmp1 = virtual2[2]
    assert isinstance(tmp1, ak._v2.contents.VirtualArray)
    assert counter == [1, 0]

    tmp2 = tmp1[1]
    assert tmp2 == 4.4
    assert counter == [1, 1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_iterator():
    content = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3]))
    offsets = ak.layout.Index32(np.array([0, 2, 2, 3], "i4"))
    array = ak.layout.ListOffsetArray32(offsets, content)
    content = v1_to_v2(content)
    array = v1_to_v2(array)

    assert list(content) == [1.1, 2.2, 3.3]
    assert [np.asarray(x).tolist() for x in array] == [[1.1, 2.2], [], [3.3]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

from awkward._ext import Index8
from awkward._ext import IndexU8
from awkward._ext import Index32
from awkward._ext import IndexU32
from awkward._ext import Index64

from awkward._ext import Identities32
from awkward._ext import Identities64

from awkward._ext import Iterator
from awkward._ext import ArrayBuilder
from awkward._ext import LayoutBuilder32
from awkward._ext import LayoutBuilder64
from awkward._ext import _PersistentSharedPtr

from awkward._ext import Content

from awkward._ext import EmptyArray

from awkward._ext import IndexedArray32
from awkward._ext import IndexedArrayU32
from awkward._ext import IndexedArray64
from awkward._ext import IndexedOptionArray32
from awkward._ext import IndexedOptionArray64

from awkward._ext import ByteMaskedArray
from awkward._ext import BitMaskedArray
from awkward._ext import UnmaskedArray

from awkward._ext import ListArray32
from awkward._ext import ListArrayU32
from awkward._ext import ListArray64

from awkward._ext import ListOffsetArray32
from awkward._ext import ListOffsetArrayU32
from awkward._ext import ListOffsetArray64

from awkward._ext import NumpyArray

from awkward._ext import Record
from awkward._ext import RecordArray

from awkward._ext import RegularArray

from awkward._ext import UnionArray8_32
from awkward._ext import UnionArray8_U32
from awkward._ext import UnionArray8_64

from awkward._ext import VirtualArray
from awkward._ext import ArrayGenerator
from awkward._ext import SliceGenerator
from awkward._ext import ArrayCache

from awkward._ext import kernel_lib


__all__ = [
    "Index8",
    "IndexU8",
    "Index32",
    "IndexU32",
    "Index64",
    "Identities32",
    "Identities64",
    "Iterator",
    "ArrayBuilder",
    "LayoutBuilder32",
    "LayoutBuilder64",
    "_PersistentSharedPtr",
    "Content",
    "EmptyArray",
    "IndexedArray32",
    "IndexedArrayU32",
    "IndexedArray64",
    "IndexedOptionArray32",
    "IndexedOptionArray64",
    "ByteMaskedArray",
    "BitMaskedArray",
    "UnmaskedArray",
    "ListArray32",
    "ListArrayU32",
    "ListArray64",
    "ListOffsetArray32",
    "ListOffsetArrayU32",
    "ListOffsetArray64",
    "NumpyArray",
    "Record",
    "RecordArray",
    "RegularArray",
    "UnionArray8_32",
    "UnionArray8_U32",
    "UnionArray8_64",
    "VirtualArray",
    "ArrayGenerator",
    "SliceGenerator",
    "ArrayCache",
    "kernel_lib",
]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert (
        np.asarray(
            ak.layout.ByteMaskedArray(
                ak.layout.Index8(np.array([1, 1, 1, 1, 1], np.int8)),
                ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
                valid_when=True,
            ).bytemask()
        ).tolist()
        == [0, 0, 0, 0, 0]
    )

    assert (
        np.asarray(
            ak.layout.ByteMaskedArray(
                ak.layout.Index8(np.array([1, 1, 1, 1, 1], np.int8)),
                ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
                valid_when=False,
            ).bytemask()
        ).tolist()
        == [1, 1, 1, 1, 1]
    )

    assert (
        np.asarray(
            ak.layout.BitMaskedArray(
                ak.layout.IndexU8(np.array([31], np.uint8)),
                ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
                valid_when=True,
                length=5,
                lsb_order=True,
            ).bytemask()
        ).tolist()
        == [0, 0, 0, 0, 0]
    )

    assert (
        np.asarray(
            ak.layout.BitMaskedArray(
                ak.layout.IndexU8(np.array([31], np.uint8)),
                ak.layout.NumpyArray(np.array([0, 1, 2, 3, 4])),
                valid_when=False,
                length=5,
                lsb_order=True,
            ).bytemask()
        ).tolist()
        == [1, 1, 1, 1, 1]
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import sys

import awkward as ak


def mixin_class(registry, name=None):
    """
    Args:
        registry (dict): The destination behavior mapping registry. Typically,
            this would be the global registry #ak.behavior, but one may wish
            to register methods in an alternative way.
        name (str): The name to assign to the behaviour class.

    This decorator can be used to register a behavior mixin class.

    Any inherited behaviors will automatically be made available to the decorated
    class.

    See the "Mixin decorators" section of #ak.behavior for further details.
    """

    def register(cls):
        cls_name = cls.__name__
        if name is None:
            behavior_name = cls_name
        else:
            behavior_name = name

        record = type(
            cls_name + "Record",
            (cls, ak.highlevel.Record),
            {"__module__": cls.__module__},
        )
        setattr(sys.modules[cls.__module__], cls_name + "Record", record)
        registry[behavior_name] = record
        array = type(
            cls_name + "Array",
            (cls, ak.highlevel.Array),
            {"__module__": cls.__module__},
        )
        setattr(sys.modules[cls.__module__], cls_name + "Array", array)
        registry["*", behavior_name] = array
        for basecls in cls.mro():
            for method in basecls.__dict__.values():
                if hasattr(method, "_awkward_mixin"):
                    ufunc, rhs, transpose = method._awkward_mixin
                    if rhs is None:
                        registry.setdefault((ufunc, behavior_name), method)
                        continue
                    for rhs_name in list(rhs) + [behavior_name]:
                        registry.setdefault((ufunc, behavior_name, rhs_name), method)
                        if transpose is not None and rhs_name != behavior_name:
                            registry.setdefault(
                                (ufunc, rhs_name, behavior_name), transpose
                            )
                    if basecls.__name__ in rhs:
                        rhs.add(behavior_name)
        return cls

    return register


def mixin_class_method(ufunc, rhs=None, transpose=True):
    """
    Args:
        ufunc (numpy.ufunc): A universal function (or NEP18 callable) that is
            hooked in Awkward Array, i.e. it can be the first argument of a behavior.
        rhs (Set[type] or None): Set of right-hand side argument types, optional
            if wrapping a unary function. The left-hand side is expected to
            always be `self` of the parent class.
        transpose (bool): If true, automatically create a transpose signature
            (only makes sense for binary ufuncs).

    This decorator can be used to register a mixin class method.

    Using this decorator ensures that derived classes that are declared with the
    #ak.mixin_class decorator will also have the behaviors that this class has.
    """

    def register(method):
        if not isinstance(rhs, (set, type(None))):
            raise ValueError(
                "expected a set of right-hand-side argument types"
                + ak._util.exception_suffix(__file__)
            )
        if transpose and rhs is not None:

            def transposed(left, right):
                return method(right, left)

            # make a copy of rhs, we will edit it later
            method._awkward_mixin = (ufunc, set(rhs), transposed)
        else:
            method._awkward_mixin = (ufunc, rhs, None)
        return method

    return register


__all__ = [
    x for x in list(globals()) if not x.startswith("_") and x not in ("sys", "ak")
]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak
from awkward._v2.record import Record
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.recordform import RecordForm

np = ak.nplike.NumpyMetadata.instance()


class RecordArray(Content):
    def __init__(self, contents, keys, length=None, identifier=None, parameters=None):
        if not isinstance(contents, Iterable):
            raise TypeError(
                "{0} 'contents' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if not isinstance(contents, list):
            contents = list(contents)

        if len(contents) == 0 and length is None:
            raise TypeError(
                "{0} if len(contents) == 0, a 'length' must be specified".format(
                    type(self).__name__
                )
            )
        elif length is None:
            length = min(len(x) for x in contents)
        if not (ak._util.isint(length) and length >= 0):
            raise TypeError(
                "{0} 'length' must be a non-negative integer or None, not {1}".format(
                    type(self).__name__, repr(length)
                )
            )
        for content in contents:
            if not isinstance(content, Content):
                raise TypeError(
                    "{0} all 'contents' must be Content subclasses, not {1}".format(
                        type(self).__name__, repr(content)
                    )
                )
            if not len(content) >= length:
                raise ValueError(
                    "{0} len(content) ({1}) must be <= length ({2}) for all 'contents'".format(
                        type(self).__name__, len(content), length
                    )
                )

        if isinstance(keys, Iterable):
            if not isinstance(keys, list):
                keys = list(keys)
            if not all(ak._util.isstr(x) for x in keys):
                raise TypeError(
                    "{0} 'keys' must all be strings, not {1}".format(
                        type(self).__name__, repr(keys)
                    )
                )
            if not len(contents) == len(keys):
                raise ValueError(
                    "{0} len(contents) ({1}) must be equal to len(keys) ({2})".format(
                        type(self).__name__, len(contents), len(keys)
                    )
                )
        elif keys is not None:
            raise TypeError(
                "{0} 'keys' must be iterable or None, not {1}".format(
                    type(self).__name__, repr(keys)
                )
            )

        self._contents = contents
        self._keys = keys
        self._length = length
        self._init(identifier, parameters)

    @property
    def contents(self):
        return self._contents

    @property
    def keys(self):
        if self._keys is None:
            return [str(i) for i in range(len(self._contents))]
        else:
            return self._keys

    @property
    def is_tuple(self):
        return self._keys is None

    @property
    def as_tuple(self):
        return RecordArray(
            self._contents,
            None,
            length=self._length,
            identifier=None,
            parameters=None,
        )

    @property
    def nplike(self):
        for content in self._contents:
            nplike = content.nonvirtual_nplike
            if nplike is not None:
                return nplike
        else:
            for content in self._contents:
                return content.nplike
            else:
                return ak.nplike.Numpy.instance()

    @property
    def nonvirtual_nplike(self):
        for content in self._contents:
            nplike = content.nonvirtual_nplike
            if nplike is not None:
                return nplike
        else:
            return None

    Form = RecordForm

    @property
    def form(self):
        return self.Form(
            [x.form for x in self._contents],
            self._keys,
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        return RecordArray(
            [x.typetracer for x in self._contents],
            self._keys,
            self._length,
            self._typetracer_identifier(),
            self._parameters,
        )

    def __len__(self):
        return self._length

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<RecordArray is_tuple="]
        out.append(repr(json.dumps(self.is_tuple)))
        out.append(" len=")
        out.append(repr(str(len(self))))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")

        if self._keys is None:
            for i, x in enumerate(self._contents):
                out.append("{0}    <content index={1}>\n".format(indent, repr(str(i))))
                out.append(x._repr(indent + "        ", "", "\n"))
                out.append(indent + "    </content>\n")
        else:
            for i, x in enumerate(self._contents):
                out.append(
                    "{0}    <content index={1} key={2}>\n".format(
                        indent, repr(str(i)), repr(self._keys[i])
                    )
                )
                out.append(x._repr(indent + "        ", "", "\n"))
                out.append(indent + "    </content>\n")

        out.append(indent + "</RecordArray>")
        out.append(post)
        return "".join(out)

    def index_to_key(self, index):
        return self.Form.index_to_key(self, index)

    def key_to_index(self, key):
        return self.Form.key_to_index(self, key)

    def haskey(self, key):
        return self.Form.haskey(self, key)

    def content(self, index_or_key):
        return self.Form.content(self, index_or_key)[: self._length]

    def _getitem_nothing(self):
        return self._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if where < 0:
            where += len(self)
        if not (0 <= where < len(self)) and self.nplike.known_shape:
            raise NestedIndexError(self, where)
        return Record(self, where)

    def _getitem_range(self, where):
        start, stop, step = where.indices(len(self))
        assert step == 1
        if len(self._contents) == 0:
            start = min(max(start, 0), self._length)
            stop = min(max(stop, 0), self._length)
            if stop < start:
                stop = start
            return RecordArray(
                [],
                self._keys,
                stop - start,
                self._range_identifier(start, stop),
                self._parameters,
            )
        else:
            nextslice = slice(start, stop)
            return RecordArray(
                [x._getitem_range(nextslice) for x in self._contents],
                self._keys,
                stop - start,
                self._range_identifier(start, stop),
                self._parameters,
            )

    def _getitem_field(self, where, only_fields=()):
        if len(only_fields) == 0:
            return self.content(where)

        else:
            nexthead, nexttail = ak._v2._slicing.headtail(only_fields)
            if ak._util.isstr(nexthead):
                return self.content(where)._getitem_field(nexthead, nexttail)
            else:
                return self.content(where)._getitem_fields(nexthead, nexttail)

    def _getitem_fields(self, where, only_fields=()):
        indexes = [self.key_to_index(key) for key in where]
        if self._keys is None:
            keys = None
        else:
            keys = [self._keys[i] for i in indexes]

        if len(only_fields) == 0:
            contents = [self.content(i) for i in indexes]
        else:
            nexthead, nexttail = ak._v2._slicing.headtail(only_fields)
            if ak._util.isstr(nexthead):
                contents = [
                    self.content(i)._getitem_field(nexthead, nexttail) for i in indexes
                ]
            else:
                contents = [
                    self.content(i)._getitem_fields(nexthead, nexttail) for i in indexes
                ]

        return RecordArray(
            contents,
            keys,
            self._fields_identifier(where),
            None,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        if allow_lazy:
            nplike, where = carry.nplike, carry.data

            if allow_lazy != "copied":
                where = where.copy()

            negative = where < 0
            if nplike.any(negative, prefer=False):
                where[negative] += self._length

            if nplike.any(where >= self._length, prefer=False):
                if issubclass(exception, NestedIndexError):
                    raise exception(self, where)
                else:
                    raise exception("index out of range")

            nextindex = ak._v2.index.Index64(where)
            return ak._v2.contents.indexedarray.IndexedArray(
                nextindex,
                self,
                self._carry_identifier(carry, exception),
                None,
            )

        else:
            contents = [
                self.content(i)._carry(carry, allow_lazy, exception)
                for i in range(len(self._contents))
            ]

            # if issubclass(carry.dtype.type, np.integer):
            #     length = len(carry)
            # else:
            #     length = len(self)
            assert issubclass(carry.dtype.type, np.integer)
            length = len(carry)

            return RecordArray(
                contents,
                self._keys,
                length,
                self._carry_identifier(carry, exception),
                self._parameters,
            )

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        contents = []
        for i in range(len(self._contents)):
            contents.append(
                self.content(i)._getitem_next_jagged(
                    slicestarts, slicestops, slicecontent, tail
                )
            )
        return RecordArray(contents, self._keys, self._length)

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            nexthead, nexttail = ak._v2._slicing.headtail(tail)

            contents = []
            for i in range(len(self._contents)):
                contents.append(self.content(i)._getitem_next(head, (), advanced))

            parameters = None
            if (
                isinstance(
                    head,
                    (
                        slice,
                        ak._v2.contents.ListOffsetArray,
                        ak._v2.contents.IndexedOptionArray,
                    ),
                )
                or head is Ellipsis
                or advanced is None
            ):
                parameters = self._parameters
            next = RecordArray(contents, self._keys, None, self._identifier, parameters)
            return next._getitem_next(nexthead, nexttail, advanced)

    def _localindex(self, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        else:
            contents = []
            for content in self._contents:
                contents.append(content._localindex(posaxis, depth))
            return RecordArray(
                contents, self._keys, len(self), self._identifier, self._parameters
            )

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self._keys is None or len(self._keys) == 0:
            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))
        else:
            raise NotImplementedError

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        if self._keys is None or len(self._keys) == 0:
            return ak._v2.contents.NumpyArray(self.nplike.instance().empty(0, np.int64))

        contents = []
        for content in self._contents:
            contents.append(
                content._sort_next(
                    negaxis,
                    starts,
                    parents,
                    outlength,
                    ascending,
                    stable,
                    kind,
                    order,
                )
            )
        return RecordArray(contents, self._keys, len(self), None, self._parameters)

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            contents = []
            for content in self._contents:
                contents.append(
                    content._combinations(
                        n, replacement, recordlookup, parameters, posaxis, depth
                    )
                )
            return ak._v2.contents.recordarray.RecordArray(
                contents, recordlookup, len(self), self._identifier, self._parameters
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        contents = []
        for content in self._contents:
            contents.append(
                content[: self._length]._reduce_next(
                    reducer,
                    negaxis,
                    starts,
                    shifts,
                    parents,
                    outlength,
                    mask,
                    keepdims,
                )
            )

        return ak._v2.contents.RecordArray(
            contents,
            self._keys,
            outlength,
            None,
            None,
        )

    def _validityerror(self, path):
        for i in range(len(self.contents)):
            if len(self.contents[i]) < len(self):
                return 'at {0} ("{1}"): len(field({2})) < len(recordarray)'.format(
                    path, type(self), i
                )
        for i in range(len(self.contents)):
            sub = self.contents[i].validityerror(path + ".field({0})".format(i))
            if sub != "":
                return sub
        return ""



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    one = ak.Array([1, 2, 3, 4])
    two = ak.Array(["aa", "bb", "cc", "dd"])
    with pytest.raises(ValueError):
        ak.cartesian([one, two], axis=1)
    with pytest.raises(ValueError):
        ak.combinations(two, 2, axis=1)

    two = ak.Array([["aa", "bb"], ["cc"], [], ["dd"]])
    assert ak.to_list(ak.cartesian([one, two], axis=1)) == [
        [(1, "aa"), (1, "bb")],
        [(2, "cc")],
        [],
        [(4, "dd")],
    ]
    assert ak.to_list(ak.combinations(two, 2, axis=1)) == [[("aa", "bb")], [], [], []]

    with pytest.raises(ValueError):
        ak.cartesian([one, two], axis=2)
    with pytest.raises(ValueError):
        ak.combinations(two, 2, axis=2)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_array_slice():
    array = ak.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    )
    array = v1_to_v2(array.layout)
    assert ak.to_list(array[[5, 2, 2, 3, 9, 0, 1]]) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        array.typetracer[[5, 2, 2, 3, 9, 0, 1]].form
        == array[[5, 2, 2, 3, 9, 0, 1]].form
    )
    assert ak.to_list(array[np.array([5, 2, 2, 3, 9, 0, 1])]) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        array.typetracer[np.array([5, 2, 2, 3, 9, 0, 1])].form
        == array[np.array([5, 2, 2, 3, 9, 0, 1])].form
    )

    array2 = ak.layout.NumpyArray(np.array([5, 2, 2, 3, 9, 0, 1], dtype=np.int32))
    array2 = v1_to_v2(array2)

    assert ak.to_list(array[array2]) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert array.typetracer[array2].form == array[array2].form
    assert ak.to_list(
        array[
            ak.Array(np.array([5, 2, 2, 3, 9, 0, 1], dtype=np.int32), check_valid=True)
        ]
    ) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert (
        array.typetracer[
            ak.Array(np.array([5, 2, 2, 3, 9, 0, 1], dtype=np.int32), check_valid=True)
        ].form
        == array[
            ak.Array(np.array([5, 2, 2, 3, 9, 0, 1], dtype=np.int32), check_valid=True)
        ].form
    )
    assert ak.to_list(array[ak.Array([5, 2, 2, 3, 9, 0, 1], check_valid=True)]) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        array.typetracer[ak.Array([5, 2, 2, 3, 9, 0, 1])].form
        == array[ak.Array([5, 2, 2, 3, 9, 0, 1])].form
    )

    array3 = ak.layout.NumpyArray(
        np.array([False, False, False, False, False, True, False, True, False, True])
    )
    array3 = v1_to_v2(array3)
    assert ak.to_list(array[array3]) == [5.5, 7.7, 9.9]
    assert array.typetracer[array3].form == array[array3].form

    content = ak.layout.NumpyArray(np.array([1, 0, 9, 3, 2, 2, 5], dtype=np.int64))
    index = ak.layout.Index64(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, content)

    indexedarray = v1_to_v2(indexedarray)

    assert ak.to_list(array[indexedarray]) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert array.typetracer[indexedarray].form == array[indexedarray].form
    assert ak.to_list(array[ak.Array(indexedarray, check_valid=True)]) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        array.typetracer[ak.Array(indexedarray)].form
        == array[ak.Array(indexedarray)].form
    )

    emptyarray = ak.layout.EmptyArray()
    emptyarray = v1_to_v2(emptyarray)

    assert ak.to_list(array[emptyarray]) == []
    assert array.typetracer[emptyarray].form == array[emptyarray].form

    content0 = ak.layout.NumpyArray(np.array([5, 2, 2]))
    content1 = ak.layout.NumpyArray(np.array([3, 9, 0, 1]))
    tags = ak.layout.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index2 = ak.layout.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.layout.UnionArray8_64(tags, index2, [content0, content1])

    unionarray = v1_to_v2(unionarray)
    assert ak.to_list(array[ak.Array(unionarray, check_valid=True)]) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        array.typetracer[ak.Array(unionarray)].form == array[ak.Array(unionarray)].form
    )

    array = ak.Array(
        np.array([[0.0, 1.1, 2.2, 3.3, 4.4], [5.5, 6.6, 7.7, 8.8, 9.9]]),
        check_valid=True,
    )

    array = v1_to_v2(array.layout)

    numpyarray1 = ak.layout.NumpyArray(np.array([[0, 1], [1, 0]]))
    numpyarray2 = ak.layout.NumpyArray(np.array([[2, 4], [3, 3]]))

    numpyarray1 = v1_to_v2(numpyarray1)
    numpyarray2 = v1_to_v2(numpyarray2)
    assert (
        ak.to_list(
            array[
                numpyarray1,
                numpyarray2,
            ]
        )
        == [[2.2, 9.9], [8.8, 3.3]]
    )
    assert (
        array.typetracer[
            numpyarray1,
            numpyarray2,
        ].form
        == array[
            numpyarray1,
            numpyarray2,
        ].form
    )
    assert ak.to_list(array[numpyarray1]) == [
        [[0.0, 1.1, 2.2, 3.3, 4.4], [5.5, 6.6, 7.7, 8.8, 9.9]],
        [[5.5, 6.6, 7.7, 8.8, 9.9], [0.0, 1.1, 2.2, 3.3, 4.4]],
    ]
    assert array.typetracer[numpyarray1].form == array[numpyarray1].form


def test_array_slice_1():
    array = ak.Array(
        [
            {"x": 1, "y": 1.1, "z": [1]},
            {"x": 2, "y": 2.2, "z": [2, 2]},
            {"x": 3, "y": 3.3, "z": [3, 3, 3]},
            {"x": 4, "y": 4.4, "z": [4, 4, 4, 4]},
            {"x": 5, "y": 5.5, "z": [5, 5, 5, 5, 5]},
        ],
        check_valid=True,
    ).layout
    array = v1_to_v2(array)
    assert ak.to_list(array[ak.from_iter(["y", "x"], highlevel=False)]) == [
        {"y": 1.1, "x": 1},
        {"y": 2.2, "x": 2},
        {"y": 3.3, "x": 3},
        {"y": 4.4, "x": 4},
        {"y": 5.5, "x": 5},
    ]


@pytest.mark.skip(reason="FIXME: UnionArray as a slice has not been implemented")
def test_array_slice_2():
    array = ak.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    )
    array = v1_to_v2(array.layout)

    content0 = ak.layout.NumpyArray(np.array([5, 2, 2]))
    content1 = ak.layout.NumpyArray(np.array([3, 9, 0, 1]))
    tags = ak.layout.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index2 = ak.layout.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.layout.UnionArray8_64(tags, index2, [content0, content1])

    unionarray = v1_to_v2(unionarray)
    assert ak.to_list(array[unionarray]) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert array.typetracer[unionarray].form == array[unionarray].form


def test_new_slices():
    content = ak.layout.NumpyArray(np.array([1, 0, 9, 3, 2, 2, 5], dtype=np.int64))
    index = ak.layout.Index64(np.array([6, 5, -1, 3, 2, -1, 0], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, content)

    indexedarray = v1_to_v2(indexedarray)
    assert ak.to_list(indexedarray) == [5, 2, None, 3, 9, None, 1]

    assert (
        ak._ext._slice_tostring(indexedarray)
        == "[missing([0, 1, -1, 2, 3, -1, 4], array([5, 2, 3, 9, 1]))]"
    )

    offsets = ak.layout.Index64(np.array([0, 4, 4, 7], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)

    listoffsetarray = v1_to_v2(listoffsetarray)
    assert ak.to_list(listoffsetarray) == [[1, 0, 9, 3], [], [2, 2, 5]]

    assert (
        ak._ext._slice_tostring(listoffsetarray)
        == "[jagged([0, 4, 4, 7], array([1, 0, 9, 3, 2, 2, 5]))]"
    )

    offsets = ak.layout.Index64(np.array([1, 4, 4, 6], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)

    listoffsetarray = v1_to_v2(listoffsetarray)
    assert ak.to_list(listoffsetarray) == [[0, 9, 3], [], [2, 2]]

    assert (
        ak._ext._slice_tostring(listoffsetarray)
        == "[jagged([0, 3, 3, 5], array([0, 9, 3, 2, 2]))]"
    )

    starts = ak.layout.Index64(np.array([1, 99, 5], dtype=np.int64))
    stops = ak.layout.Index64(np.array([4, 99, 7], dtype=np.int64))
    listarray = ak.layout.ListArray64(starts, stops, content)

    listarray = v1_to_v2(listarray)
    assert ak.to_list(listarray) == [[0, 9, 3], [], [2, 5]]

    assert (
        ak._ext._slice_tostring(listarray)
        == "[jagged([0, 3, 3, 5], array([0, 9, 3, 2, 5]))]"
    )


def test_missing():
    array = ak.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    )
    array = v1_to_v2(array.layout)
    array2 = ak.Array([3, 6, None, None, -2, 6], check_valid=True)
    array2 = v1_to_v2(array2.layout)
    assert ak.to_list(array[array2]) == [
        3.3,
        6.6,
        None,
        None,
        8.8,
        6.6,
    ]
    assert array.typetracer[array2].form == array[array2].form

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0, 11.1, 999])
    )
    regulararray = ak.layout.RegularArray(content, 4, zeros_length=0)

    regulararray = v1_to_v2(regulararray)
    assert ak.to_list(regulararray) == [
        [0.0, 1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6, 7.7],
        [8.8, 9.9, 10.0, 11.1],
    ]
    array3 = ak.Array([2, 1, 1, None, -1], check_valid=True)
    array3 = v1_to_v2(array3.layout)
    assert ak.to_list(regulararray[array3]) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert regulararray.typetracer[array3].form == regulararray[array3].form
    assert ak.to_list(regulararray[:, array3]) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert regulararray.typetracer[:, array3].form == regulararray[:, array3].form
    assert ak.to_list(regulararray[1:, array3]) == [
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert regulararray.typetracer[1:, array3].form == regulararray[1:, array3].form

    assert ak.to_list(
        regulararray[
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False])
        ]
    ) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert (
        regulararray.typetracer[
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False])
        ].form
        == regulararray[
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False])
        ].form
    )
    assert ak.to_list(
        regulararray[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ]
    ) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        regulararray.typetracer[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
        == regulararray[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
    )

    assert (
        ak.to_list(
            regulararray[
                1:,
                np.ma.MaskedArray(
                    [2, 1, 1, 999, -1], [False, False, False, True, False]
                ),
            ]
        )
        == [[6.6, 5.5, 5.5, None, 7.7], [10.0, 9.9, 9.9, None, 11.1]]
    )
    assert (
        regulararray.typetracer[
            1:,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
        == regulararray[
            1:,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
    )

    content = ak.layout.NumpyArray(
        np.array([[0.0, 1.1, 2.2, 3.3], [4.4, 5.5, 6.6, 7.7], [8.8, 9.9, 10.0, 11.1]])
    )
    content = v1_to_v2(content)
    assert ak.to_list(content[array3]) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert content.typetracer[array3].form == content[array3].form
    assert ak.to_list(content[:, array3]) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert content.typetracer[:, array3].form == content[:, array3].form
    assert ak.to_list(content[1:, array3]) == [
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert content.typetracer[1:, array3].form == content[1:, array3].form

    assert ak.to_list(
        content[
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False])
        ]
    ) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert (
        content.typetracer[
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False])
        ].form
        == content[
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False])
        ].form
    )
    assert ak.to_list(
        content[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ]
    ) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        content.typetracer[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
        == content[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
    )
    assert (
        ak.to_list(
            content[
                1:,
                np.ma.MaskedArray(
                    [2, 1, 1, 999, -1], [False, False, False, True, False]
                ),
            ]
        )
        == [[6.6, 5.5, 5.5, None, 7.7], [10.0, 9.9, 9.9, None, 11.1]]
    )
    assert (
        content.typetracer[
            1:,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
        == content[
            1:,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
    )

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0, 11.1, 999])
    )
    offsets = ak.layout.Index64(np.array([0, 4, 8, 12], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)

    listoffsetarray = v1_to_v2(listoffsetarray)
    assert ak.to_list(listoffsetarray) == [
        [0.0, 1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6, 7.7],
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert ak.to_list(listoffsetarray[:, array3]) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert listoffsetarray.typetracer[:, array3].form == listoffsetarray[:, array3].form
    assert ak.to_list(listoffsetarray[1:, array3]) == [
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        listoffsetarray.typetracer[1:, array3].form == listoffsetarray[1:, array3].form
    )

    assert ak.to_list(
        listoffsetarray[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ]
    ) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        listoffsetarray.typetracer[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
        == listoffsetarray[
            :,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
    )
    assert (
        ak.to_list(
            listoffsetarray[
                1:,
                np.ma.MaskedArray(
                    [2, 1, 1, 999, -1], [False, False, False, True, False]
                ),
            ]
        )
        == [[6.6, 5.5, 5.5, None, 7.7], [10.0, 9.9, 9.9, None, 11.1]]
    )
    assert (
        listoffsetarray.typetracer[
            1:,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
        == listoffsetarray[
            1:,
            np.ma.MaskedArray([2, 1, 1, 999, -1], [False, False, False, True, False]),
        ].form
    )


def test_bool_missing():
    data = [1.1, 2.2, 3.3, 4.4, 5.5]
    array = ak.layout.NumpyArray(np.array(data))

    array = v1_to_v2(array)

    assert (
        ak._ext._slice_tostring(
            ak.Array([True, False, None, True, False], check_valid=True)
        )
        == "[missing([0, -1, 1], array([0, 3]))]"
    )
    assert (
        ak._ext._slice_tostring(ak.Array([None, None, None], check_valid=True))
        == "[missing([-1, -1, -1], array([]))]"
    )

    x1, x2, x3, x4, x5 = True, True, True, False, None
    mask = [x1, x2, x3, x4, x5]
    expected = [m if m is None else x for x, m in zip(data, mask) if m is not False]
    array2 = ak.Array(mask, check_valid=True)
    array2 = v1_to_v2(array2.layout)

    for x1 in [True, False, None]:
        for x2 in [True, False, None]:
            for x3 in [True, False, None]:
                for x4 in [True, False, None]:
                    for x5 in [True, False, None]:
                        mask = [x1, x2, x3, x4, x5]
                        expected = [
                            m if m is None else x
                            for x, m in zip(data, mask)
                            if m is not False
                        ]
                        array2 = ak.Array(mask, check_valid=True)
                        array2 = v1_to_v2(array2.layout)
                        assert ak.to_list(array[array2]) == expected
                        assert array.typetracer[array2].form == array[array2].form


def test_bool_missing2():
    array = ak.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    )
    array2 = ak.Array([3, 6, None, None, -2, 6], check_valid=True)

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array[array2]) == [
        3.3,
        6.6,
        None,
        None,
        8.8,
        6.6,
    ]
    assert array.typetracer[array2].form == array[array2].form

    array = ak.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    array1 = v1_to_v2(array)

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0, 11.1, 999])
    )
    regulararray = ak.layout.RegularArray(content, 4, zeros_length=0)

    assert ak.to_list(regulararray) == [
        [0.0, 1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6, 7.7],
        [8.8, 9.9, 10.0, 11.1],
    ]

    regulararray = v1_to_v2(regulararray)

    array1 = v1_to_v2(ak.from_iter([True, None, False, True], highlevel=False))

    assert ak.to_list(regulararray[:, array1]) == [
        [0.0, None, 3.3],
        [4.4, None, 7.7],
        [8.8, None, 11.1],
    ]
    assert regulararray.typetracer[:, array1].form == regulararray[:, array1].form

    assert ak.to_list(regulararray[1:, array1]) == [
        [4.4, None, 7.7],
        [8.8, None, 11.1],
    ]
    assert regulararray.typetracer[1:, array1].form == regulararray[1:, array1].form

    content = ak.layout.NumpyArray(
        np.array([[0.0, 1.1, 2.2, 3.3], [4.4, 5.5, 6.6, 7.7], [8.8, 9.9, 10.0, 11.1]])
    )
    content = v1_to_v2(content)

    assert ak.to_list(content[:, array1]) == [
        [0.0, None, 3.3],
        [4.4, None, 7.7],
        [8.8, None, 11.1],
    ]
    assert content.typetracer[:, array1].form == content[:, array1].form

    assert ak.to_list(content[1:, array1]) == [[4.4, None, 7.7], [8.8, None, 11.1]]
    assert content.typetracer[1:, array1].form == content[1:, array1].form

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0, 11.1, 999])
    )
    offsets = ak.layout.Index64(np.array([0, 4, 8, 12], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)

    listoffsetarray = v1_to_v2(listoffsetarray)

    assert ak.to_list(listoffsetarray[:, array1]) == [
        [0.0, None, 3.3],
        [4.4, None, 7.7],
        [8.8, None, 11.1],
    ]
    assert listoffsetarray.typetracer[:, array1].form == listoffsetarray[:, array1].form

    assert ak.to_list(listoffsetarray[1:, array1]) == [
        [4.4, None, 7.7],
        [8.8, None, 11.1],
    ]
    assert (
        listoffsetarray.typetracer[1:, array1].form == listoffsetarray[1:, array1].form
    )


def test_records_missing():
    array = ak.Array(
        [
            {"x": 0, "y": 0.0},
            {"x": 1, "y": 1.1},
            {"x": 2, "y": 2.2},
            {"x": 3, "y": 3.3},
            {"x": 4, "y": 4.4},
            {"x": 5, "y": 5.5},
            {"x": 6, "y": 6.6},
            {"x": 7, "y": 7.7},
            {"x": 8, "y": 8.8},
            {"x": 9, "y": 9.9},
        ],
        check_valid=True,
    )
    array2 = ak.Array([3, 1, None, 1, 7], check_valid=True)

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array[array2]) == [
        {"x": 3, "y": 3.3},
        {"x": 1, "y": 1.1},
        None,
        {"x": 1, "y": 1.1},
        {"x": 7, "y": 7.7},
    ]
    assert array.typetracer[array2].form == array[array2].form

    array = ak.Array(
        [
            [
                {"x": 0, "y": 0.0},
                {"x": 1, "y": 1.1},
                {"x": 2, "y": 2.2},
                {"x": 3, "y": 3.3},
            ],
            [
                {"x": 4, "y": 4.4},
                {"x": 5, "y": 5.5},
                {"x": 6, "y": 6.6},
                {"x": 7, "y": 7.7},
                {"x": 8, "y": 8.8},
                {"x": 9, "y": 9.9},
            ],
        ],
        check_valid=True,
    )
    array2 = ak.Array([1, None, 2, -1], check_valid=True)

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array[:, array2]) == [
        [{"x": 1, "y": 1.1}, None, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        [{"x": 5, "y": 5.5}, None, {"x": 6, "y": 6.6}, {"x": 9, "y": 9.9}],
    ]
    assert array.typetracer[:, array2].form == array[:, array2].form

    array = ak.Array(
        [
            {"x": [0, 1, 2, 3], "y": [0.0, 1.1, 2.2, 3.3]},
            {"x": [4, 5, 6, 7], "y": [4.4, 5.5, 6.6, 7.7]},
            {"x": [8, 9, 10, 11], "y": [8.8, 9.9, 10.0, 11.1]},
        ],
        check_valid=True,
    )

    array = v1_to_v2(array.layout)

    assert ak.to_list(array[:, array2]) == [
        {"x": [1, None, 2, 3], "y": [1.1, None, 2.2, 3.3]},
        {"x": [5, None, 6, 7], "y": [5.5, None, 6.6, 7.7]},
        {"x": [9, None, 10, 11], "y": [9.9, None, 10.0, 11.1]},
    ]
    assert array.typetracer[:, array2].form == array[:, array2].form
    assert ak.to_list(array[1:, array2]) == [
        {"x": [5, None, 6, 7], "y": [5.5, None, 6.6, 7.7]},
        {"x": [9, None, 10, 11], "y": [9.9, None, 10.0, 11.1]},
    ]
    assert array.typetracer[1:, array2].form == array[1:, array2].form


def test_jagged():
    array = ak.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    )
    array2 = ak.Array([[0, -1], [], [-1, 0], [-1], [1, 1, -2, 0]], check_valid=True)

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array[array2]) == [
        [1.1, 3.3],
        [],
        [5.5, 4.4],
        [6.6],
        [8.8, 8.8, 8.8, 7.7],
    ]
    assert array.typetracer[array2].form == array[array2].form


def test_double_jagged():
    array = ak.Array(
        [[[0, 1, 2, 3], [4, 5]], [[6, 7, 8], [9, 10, 11, 12, 13]]], check_valid=True
    )
    array2 = ak.Array(
        [[[2, 1, 0], [-1]], [[-1, -2, -3], [2, 1, 1, 3]]], check_valid=True
    )

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array[array2]) == [
        [[2, 1, 0], [5]],
        [[8, 7, 6], [11, 10, 10, 12]],
    ]
    assert array.typetracer[array2].form == array[array2].form

    content = ak.from_iter(
        [[0, 1, 2, 3], [4, 5], [6, 7, 8], [9, 10, 11, 12, 13]], highlevel=False
    )
    regulararray = ak.layout.RegularArray(content, 2, zeros_length=0)
    regulararray = v1_to_v2(regulararray)

    array1 = ak.Array([[2, 1, 0], [-1]], check_valid=True)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(regulararray[:, array1]) == [[[2, 1, 0], [5]], [[8, 7, 6], [13]]]
    assert regulararray.typetracer[:, array1].form == regulararray[:, array1].form
    assert ak.to_list(regulararray[1:, array1]) == [[[8, 7, 6], [13]]]
    assert regulararray.typetracer[1:, array1].form == regulararray[1:, array1].form

    offsets = ak.layout.Index64(np.array([0, 2, 4], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    listoffsetarray = v1_to_v2(listoffsetarray)
    assert ak.to_list(listoffsetarray[:, array1]) == [
        [[2, 1, 0], [5]],
        [[8, 7, 6], [13]],
    ]
    assert listoffsetarray.typetracer[:, array1].form == listoffsetarray[:, array1].form
    assert ak.to_list(listoffsetarray[1:, array1]) == [[[8, 7, 6], [13]]]
    assert (
        listoffsetarray.typetracer[1:, array1].form == listoffsetarray[1:, array1].form
    )


def test_masked_jagged():
    array = ak.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    )
    array1 = ak.Array([[-1, -2], None, [], None, [-2, 0]], check_valid=True)

    array = v1_to_v2(array.layout)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(array[array1]) == [[3.3, 2.2], None, [], None, [8.8, 7.7]]
    assert array.typetracer[array1].form == array[array1].form


def test_jagged_masked():
    array = ak.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    )
    array1 = ak.Array([[-1, None], [], [None, 0], [None], [1]], check_valid=True)

    array = v1_to_v2(array.layout)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(array[array1]) == [[3.3, None], [], [None, 4.4], [None], [8.8]]
    assert array.typetracer[array1].form == array[array1].form


def test_regular_regular():
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5))
    regulararray1 = ak.layout.RegularArray(content, 5, zeros_length=0)
    regulararray2 = ak.layout.RegularArray(regulararray1, 3, zeros_length=0)

    regulararray2 = v1_to_v2(regulararray2)
    array1 = ak.Array(
        [[[2], [1, -2], [-1, 2, 0]], [[-3], [-4, 3], [-5, -3, 4]]],
        check_valid=True,
    )
    array2 = ak.Array(
        [[[2], [1, -2], [-1, None, 0]], [[-3], [-4, 3], [-5, None, 4]]],
        check_valid=True,
    )
    array1 = v1_to_v2(array1.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(regulararray2[array1]) == [
        [[2], [6, 8], [14, 12, 10]],
        [[17], [21, 23], [25, 27, 29]],
    ]
    assert regulararray2.typetracer[array1].form == regulararray2[array1].form

    assert ak.to_list(regulararray2[array2]) == [
        [[2], [6, 8], [14, None, 10]],
        [[17], [21, 23], [25, None, 29]],
    ]
    assert regulararray2.typetracer[array2].form == regulararray2[array2].form


def test_masked_of_jagged_of_whatever():
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5))
    regulararray1 = ak.layout.RegularArray(content, 5, zeros_length=0)
    regulararray2 = ak.layout.RegularArray(regulararray1, 3, zeros_length=0)

    regulararray2 = v1_to_v2(regulararray2)
    array1 = ak.Array(
        [[[2], None, [-1, 2, 0]], [[-3], None, [-5, -3, 4]]], check_valid=True
    )
    array2 = ak.Array(
        [[[2], None, [-1, None, 0]], [[-3], None, [-5, None, 4]]],
        check_valid=True,
    )
    array1 = v1_to_v2(array1.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(regulararray2[array1]) == [
        [[2], None, [14, 12, 10]],
        [[17], None, [25, 27, 29]],
    ]
    assert regulararray2.typetracer[array1].form == regulararray2[array1].form

    assert ak.to_list(regulararray2[array2]) == [
        [[2], None, [14, None, 10]],
        [[17], None, [25, None, 29]],
    ]
    assert regulararray2.typetracer[array2].form == regulararray2[array2].form


def test_emptyarray():
    content = ak.layout.EmptyArray()
    offsets = ak.layout.Index64(np.array([0, 0, 0, 0, 0], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)

    listoffsetarray = v1_to_v2(listoffsetarray)
    array1 = ak.Array([[], [], [], []], check_valid=True)
    array2 = ak.Array([[], [None], [], []], check_valid=True)
    array3 = ak.Array([[], [], None, []], check_valid=True)
    array4 = ak.Array([[], [None], None, []], check_valid=True)
    array5 = ak.Array([[], [0], [], []], check_valid=True)

    array1 = v1_to_v2(array1.layout)
    array2 = v1_to_v2(array2.layout)
    array3 = v1_to_v2(array3.layout)
    array4 = v1_to_v2(array4.layout)
    array5 = v1_to_v2(array5.layout)
    assert ak.to_list(listoffsetarray) == [[], [], [], []]

    assert ak.to_list(listoffsetarray[array1]) == [[], [], [], []]
    assert listoffsetarray.typetracer[array1].form == listoffsetarray[array1].form

    assert ak.to_list(listoffsetarray[array2]) == [[], [None], [], []]
    assert listoffsetarray.typetracer[array2].form == listoffsetarray[array2].form
    assert ak.to_list(listoffsetarray[array3]) == [[], [], None, []]
    assert listoffsetarray.typetracer[array3].form == listoffsetarray[array3].form
    assert ak.to_list(listoffsetarray[array4]) == [[], [None], None, []]
    assert listoffsetarray.typetracer[array4].form == listoffsetarray[array4].form

    with pytest.raises(ValueError):
        listoffsetarray[array5]


def test_numpyarray():
    array = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)
    array2 = ak.Array([[[], [], []], [], [[], []]], check_valid=True)

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)
    with pytest.raises(IndexError):
        array[array2]


def test_record():
    array = ak.Array(
        [
            {"x": [0, 1, 2], "y": [0.0, 1.1, 2.2, 3.3]},
            {"x": [3, 4, 5, 6], "y": [4.4, 5.5]},
            {"x": [7, 8], "y": [6.6, 7.7, 8.8, 9.9]},
        ],
        check_valid=True,
    )
    array2 = ak.Array([[-1, 1], [0, 0, 1], [-1, -2]], check_valid=True)
    array3 = ak.Array([[-1, 1], [0, 0, None, 1], [-1, -2]], check_valid=True)
    array4 = ak.Array([[-1, 1], None, [-1, -2]], check_valid=True)

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)
    array3 = v1_to_v2(array3.layout)
    array4 = v1_to_v2(array4.layout)

    assert ak.to_list(array[array2]) == [
        {"x": [2, 1], "y": [3.3, 1.1]},
        {"x": [3, 3, 4], "y": [4.4, 4.4, 5.5]},
        {"x": [8, 7], "y": [9.9, 8.8]},
    ]
    assert array.typetracer[array2].form == array[array2].form
    assert ak.to_list(array[array3]) == [
        {"x": [2, 1], "y": [3.3, 1.1]},
        {"x": [3, 3, None, 4], "y": [4.4, 4.4, None, 5.5]},
        {"x": [8, 7], "y": [9.9, 8.8]},
    ]
    assert array.typetracer[array3].form == array[array3].form
    assert ak.to_list(array[array4]) == [
        {"x": [2, 1], "y": [3.3, 1.1]},
        None,
        {"x": [8, 7], "y": [9.9, 8.8]},
    ]
    assert array.typetracer[array4].form == array[array4].form


def test_indexedarray():
    array = ak.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    index = ak.layout.Index64(np.array([3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, array)

    indexedarray = v1_to_v2(indexedarray)

    assert ak.to_list(indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [0.0, 1.1, 2.2],
    ]

    array1 = ak.Array([[0, -1], [0], [], [1, 1]], check_valid=True)
    array1 = v1_to_v2(array1.layout)
    assert ak.to_list(indexedarray[array1]) == [[6.6, 9.9], [5.5], [], [1.1, 1.1]]
    assert indexedarray.typetracer[array1].form == indexedarray[array1].form

    array1 = ak.Array([[0, -1], [0], [None], [1, None, 1]], check_valid=True)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(indexedarray[array1]) == [
        [6.6, 9.9],
        [5.5],
        [None],
        [1.1, None, 1.1],
    ]
    assert indexedarray.typetracer[array1].form == indexedarray[array1].form

    array1 = ak.Array([[0, -1], [0], None, [1, 1]], check_valid=True)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(indexedarray[array1]) == [[6.6, 9.9], [5.5], None, [1.1, 1.1]]
    assert indexedarray.typetracer[array1].form == indexedarray[array1].form

    array1 = ak.Array([[0, -1], [0], None, [None]], check_valid=True)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(indexedarray[array1]) == [[6.6, 9.9], [5.5], None, [None]]
    assert indexedarray.typetracer[array1].form == indexedarray[array1].form

    index = ak.layout.Index64(np.array([3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, array)

    indexedarray = v1_to_v2(indexedarray)

    assert ak.to_list(indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [0.0, 1.1, 2.2],
    ]

    assert ak.to_list(
        indexedarray[ak.Array([[0, -1], [0], [], [1, 1]], check_valid=True)]
    ) == [[6.6, 9.9], [5.5], [], [1.1, 1.1]]
    assert (
        indexedarray.typetracer[ak.Array([[0, -1], [0], [], [1, 1]])].form
        == indexedarray[ak.Array([[0, -1], [0], [], [1, 1]])].form
    )

    array1 = ak.Array([[0, -1], [0], [None], [1, None, 1]], check_valid=True)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(indexedarray[array1]) == [
        [6.6, 9.9],
        [5.5],
        [None],
        [1.1, None, 1.1],
    ]
    assert indexedarray.typetracer[array1].form == indexedarray[array1].form

    array1 = ak.Array([[0, -1], [0], None, []], check_valid=True)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(indexedarray[array1]) == [[6.6, 9.9], [5.5], None, []]
    assert indexedarray.typetracer[array1].form == indexedarray[array1].form

    array1 = ak.Array([[0, -1], [0], None, [1, None, 1]], check_valid=True)
    array1 = v1_to_v2(array1.layout)

    assert ak.to_list(indexedarray[array1]) == [
        [6.6, 9.9],
        [5.5],
        None,
        [1.1, None, 1.1],
    ]
    assert indexedarray.typetracer[array1].form == indexedarray[array1].form


def test_indexedarray2():
    array = ak.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    index = ak.layout.Index64(np.array([3, 2, -1, 0], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, array)
    array = ak.Array([[0, -1], [0], None, [1, 1]])

    indexedarray = v1_to_v2(indexedarray)
    array = v1_to_v2(array.layout)

    assert ak.to_list(indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        None,
        [0.0, 1.1, 2.2],
    ]
    assert ak.to_list(indexedarray[array]) == [
        [6.6, 9.9],
        [5.5],
        None,
        [1.1, 1.1],
    ]
    assert indexedarray.typetracer[array].form == indexedarray[array].form


def test_indexedarray2b():
    array = ak.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    index = ak.layout.Index64(np.array([0, -1, 2, 3], dtype=np.int64))
    indexedarray = ak.layout.IndexedOptionArray64(index, array)
    array = ak.Array([[1, 1], None, [0], [0, -1]])

    indexedarray = v1_to_v2(indexedarray)
    array = v1_to_v2(array.layout)

    assert ak.to_list(indexedarray) == [
        [0.0, 1.1, 2.2],
        None,
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.to_list(indexedarray[array]) == [
        [1.1, 1.1],
        None,
        [5.5],
        [6.6, 9.9],
    ]
    assert indexedarray.typetracer[array].form == indexedarray[array].form


def test_bytemaskedarray2b():
    array = ak.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.layout.Index8(np.array([0, 1, 0, 0], dtype=np.int8))
    maskedarray = ak.layout.ByteMaskedArray(mask, array, valid_when=False)
    array = ak.Array([[1, 1], None, [0], [0, -1]])

    maskedarray = v1_to_v2(maskedarray)
    array = v1_to_v2(array.layout)

    assert ak.to_list(maskedarray) == [
        [0.0, 1.1, 2.2],
        None,
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.to_list(maskedarray[array]) == [
        [1.1, 1.1],
        None,
        [5.5],
        [6.6, 9.9],
    ]
    assert maskedarray.typetracer[array].form == maskedarray[array].form


def test_bitmaskedarray2b():
    array = ak.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.layout.IndexU8(np.array([66], dtype=np.uint8))
    maskedarray = ak.layout.BitMaskedArray(
        mask, array, valid_when=False, length=4, lsb_order=True
    )  # lsb_order is irrelevant in this example
    array = ak.Array([[1, 1], None, [0], [0, -1]])
    maskedarray = v1_to_v2(maskedarray)
    array = v1_to_v2(array.layout)

    assert ak.to_list(maskedarray) == [
        [0.0, 1.1, 2.2],
        None,
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.to_list(maskedarray[array]) == [
        [1.1, 1.1],
        None,
        [5.5],
        [6.6, 9.9],
    ]
    assert maskedarray.typetracer[array].form == maskedarray[array].form


def test_indexedarray3():
    array = ak.Array([0.0, 1.1, 2.2, None, 4.4, None, None, 7.7])
    assert ak.to_list(array[ak.Array([4, 3, 2])]) == [4.4, None, 2.2]
    assert ak.to_list(array[ak.Array([4, 3, 2, None, 1])]) == [
        4.4,
        None,
        2.2,
        None,
        1.1,
    ]

    array = ak.Array([[0.0, 1.1, None, 2.2], [3.3, None, 4.4], [5.5]])
    array2 = ak.Array([[3, 2, 2, 1], [1, 2], []])

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array[array2]) == [
        [2.2, None, None, 1.1],
        [None, 4.4],
        [],
    ]
    assert array.typetracer[array2].form == array[array2].form

    array = ak.Array([[0.0, 1.1, 2.2], [3.3, 4.4], None, [5.5]])
    array2 = ak.Array([3, 2, 1])
    array3 = ak.Array([3, 2, 1, None, 0])
    array4 = ak.Array([[2, 1, 1, 0], [1], None, [0]])
    array5 = ak.Array([[2, 1, 1, 0], None, [1], [0]])
    array6 = ak.Array([[2, 1, 1, 0], None, [1], [0], None])

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)
    array3 = v1_to_v2(array3.layout)
    array4 = v1_to_v2(array4.layout)
    array5 = v1_to_v2(array5.layout)
    array6 = v1_to_v2(array6.layout)

    assert ak.to_list(array[array2]) == [[5.5], None, [3.3, 4.4]]
    assert array.typetracer[array2].form == array[array2].form
    assert ak.to_list(array[array3]) == [
        [5.5],
        None,
        [3.3, 4.4],
        None,
        [0.0, 1.1, 2.2],
    ]
    assert array.typetracer[array3].form == array[array3].form

    assert (ak.to_list(array[array4])) == [
        [2.2, 1.1, 1.1, 0.0],
        [4.4],
        None,
        [5.5],
    ]
    assert array.typetracer[array4].form == array[array4].form

    assert ak.to_list(array[array5]) == [
        [2.2, 1.1, 1.1, 0],
        None,
        None,
        [5.5],
    ]
    assert array.typetracer[array5].form == array[array5].form
    with pytest.raises(IndexError):
        array[array6]


def test_sequential():
    array = ak.Array(np.arange(2 * 3 * 5).reshape(2, 3, 5).tolist(), check_valid=True)
    array2 = ak.Array([[2, 1, 0], [2, 1, 0]], check_valid=True)

    array = v1_to_v2(array.layout)
    array2 = v1_to_v2(array2.layout)

    assert ak.to_list(array[array2]) == [
        [[10, 11, 12, 13, 14], [5, 6, 7, 8, 9], [0, 1, 2, 3, 4]],
        [[25, 26, 27, 28, 29], [20, 21, 22, 23, 24], [15, 16, 17, 18, 19]],
    ]
    assert array.typetracer[array2].form == array[array2].form
    assert ak.to_list(array[array2, :2]) == [
        [[10, 11], [5, 6], [0, 1]],
        [[25, 26], [20, 21], [15, 16]],
    ]
    assert array.typetracer[array2, :2].form == array[array2, :2].form


def test_union():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter(
        [[6.6], [7.7, 8.8], [], [9.9, 10.0, 11.1, 12.2]], highlevel=False
    )
    tags = ak.layout.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index = ak.layout.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.layout.UnionArray8_64(tags, index, [one, two])
    array = ak.Array([[0, -1], [], [1, 1], [], [-1], [], [1, -2, -1]], check_valid=True)

    unionarray = v1_to_v2(unionarray)
    array = v1_to_v2(array.layout)
    assert ak.to_list(unionarray) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8],
        [],
        [9.9, 10.0, 11.1, 12.2],
    ]


@pytest.mark.skip(reason="FIXME: simplify_uniontype needs to be implemented")
def test_union_2():
    one = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.from_iter(
        [[6.6], [7.7, 8.8], [], [9.9, 10.0, 11.1, 12.2]], highlevel=False
    )
    tags = ak.layout.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index = ak.layout.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.layout.UnionArray8_64(tags, index, [one, two])
    array = ak.Array([[0, -1], [], [1, 1], [], [-1], [], [1, -2, -1]], check_valid=True)

    unionarray = v1_to_v2(unionarray)
    array = v1_to_v2(array.layout)
    assert ak.to_list(unionarray[array]) == [
        [1.1, 3.3],
        [],
        [5.5, 5.5],
        [],
        [8.8],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert unionarray.typetracer[array].form == unionarray[array].form


def test_jagged_mask():
    array = ak.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    )
    array = v1_to_v2(array.layout)
    assert ak.to_list(
        array[[[True, True, True], [], [True, True], [True], [True, True, True]]]
    ) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        array.typetracer[
            [[True, True, True], [], [True, True], [True], [True, True, True]]
        ].form
        == array[
            [[True, True, True], [], [True, True], [True], [True, True, True]]
        ].form
    )
    assert ak.to_list(
        array[[[False, True, True], [], [True, True], [True], [True, True, True]]]
    ) == [[2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        array.typetracer[
            [[False, True, True], [], [True, True], [True], [True, True, True]]
        ].form
        == array[
            [[False, True, True], [], [True, True], [True], [True, True, True]]
        ].form
    )
    assert ak.to_list(
        array[[[True, False, True], [], [True, True], [True], [True, True, True]]]
    ) == [[1.1, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        array.typetracer[
            [[True, False, True], [], [True, True], [True], [True, True, True]]
        ].form
        == array[
            [[True, False, True], [], [True, True], [True], [True, True, True]]
        ].form
    )
    assert ak.to_list(
        array[[[True, True, True], [], [False, True], [True], [True, True, True]]]
    ) == [[1.1, 2.2, 3.3], [], [5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        array.typetracer[
            [[True, True, True], [], [False, True], [True], [True, True, True]]
        ].form
        == array[
            [[True, True, True], [], [False, True], [True], [True, True, True]]
        ].form
    )
    assert ak.to_list(
        array[[[True, True, True], [], [False, False], [True], [True, True, True]]]
    ) == [[1.1, 2.2, 3.3], [], [], [6.6], [7.7, 8.8, 9.9]]
    assert (
        array.typetracer[
            [[True, True, True], [], [False, False], [True], [True, True, True]]
        ].form
        == array[
            [[True, True, True], [], [False, False], [True], [True, True, True]]
        ].form
    )


def test_jagged_missing_mask():
    array = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)
    array = v1_to_v2(array.layout)

    assert ak.to_list(array[[[True, True, True], [], [True, True]]]) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert (
        array.typetracer[[[True, True, True], [], [True, True]]].form
        == array[[[True, True, True], [], [True, True]]].form
    )
    assert ak.to_list(array[[[True, False, True], [], [True, True]]]) == [
        [1.1, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert (
        array.typetracer[[[True, False, True], [], [True, True]]].form
        == array[[[True, False, True], [], [True, True]]].form
    )
    assert ak.to_list(array[[[True, True, False], [], [False, None]]]) == [
        [1.1, 2.2],
        [],
        [None],
    ]
    assert (
        array.typetracer[[[True, True, False], [], [False, None]]].form
        == array[[[True, True, False], [], [False, None]]].form
    )
    assert ak.to_list(array[[[True, True, False], [], [True, None]]]) == [
        [1.1, 2.2],
        [],
        [4.4, None],
    ]
    assert (
        array.typetracer[[[True, True, False], [], [True, None]]].form
        == array[[[True, True, False], [], [True, None]]].form
    )

    assert ak.to_list(array[[[True, None, True], [], [True, True]]]) == [
        [1.1, None, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert (
        array.typetracer[[[True, None, True], [], [True, True]]].form
        == array[[[True, None, True], [], [True, True]]].form
    )
    assert ak.to_list(array[[[True, None, False], [], [True, True]]]) == [
        [1.1, None],
        [],
        [4.4, 5.5],
    ]
    assert (
        array.typetracer[[[True, None, False], [], [True, True]]].form
        == array[[[True, None, False], [], [True, True]]].form
    )

    assert ak.to_list(array[[[False, None, False], [], [True, True]]]) == [
        [None],
        [],
        [4.4, 5.5],
    ]
    assert (
        array.typetracer[[[False, None, False], [], [True, True]]].form
        == array[[[False, None, False], [], [True, True]]].form
    )
    assert ak.to_list(array[[[True, True, False], [], [False, True]]]) == [
        [1.1, 2.2],
        [],
        [5.5],
    ]
    assert (
        array.typetracer[[[True, True, False], [], [False, True]]].form
        == array[[[True, True, False], [], [False, True]]].form
    )
    assert ak.to_list(array[[[True, True, None], [], [False, True]]]) == [
        [1.1, 2.2, None],
        [],
        [5.5],
    ]
    assert (
        array.typetracer[[[True, True, None], [], [False, True]]].form
        == array[[[True, True, None], [], [False, True]]].form
    )
    assert ak.to_list(array[[[True, True, False], [None], [False, True]]]) == [
        [1.1, 2.2],
        [None],
        [5.5],
    ]
    assert (
        array.typetracer[[[True, True, False], [None], [False, True]]].form
        == array[[[True, True, False], [None], [False, True]]].form
    )


def test_array_boolean_to_int():
    a = v1_to_v2(
        ak.from_iter(
            [[True, True, True], [], [True, True], [True], [True, True, True, True]],
            highlevel=False,
        )
    )
    b = ak._v2._slicing.prepare_tuple_bool_to_int(a)
    assert ak.to_list(b) == [[0, 1, 2], [], [0, 1], [0], [0, 1, 2, 3]]

    a = v1_to_v2(
        ak.from_iter(
            [
                [True, True, False],
                [],
                [True, False],
                [False],
                [True, True, True, False],
            ],
            highlevel=False,
        )
    )
    b = ak._v2._slicing.prepare_tuple_bool_to_int(a)
    assert ak.to_list(b) == [[0, 1], [], [0], [], [0, 1, 2]]

    a = v1_to_v2(
        ak.from_iter(
            [
                [False, True, True],
                [],
                [False, True],
                [False],
                [False, True, True, True],
            ],
            highlevel=False,
        )
    )
    b = ak._v2._slicing.prepare_tuple_bool_to_int(a)
    assert ak.to_list(b) == [[1, 2], [], [1], [], [1, 2, 3]]

    a = v1_to_v2(
        ak.from_iter(
            [[True, True, None], [], [True, None], [None], [True, True, True, None]],
            highlevel=False,
        )
    )
    b = ak._v2._slicing.prepare_tuple_bool_to_int(a)
    assert ak.to_list(b) == [[0, 1, None], [], [0, None], [None], [0, 1, 2, None]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(6).tolist()  # kernels expect nonnegative entries to be arange
    )

    a = v1_to_v2(
        ak.from_iter(
            [[None, True, True], [], [None, True], [None], [None, True, True, True]],
            highlevel=False,
        )
    )
    b = ak._v2._slicing.prepare_tuple_bool_to_int(a)
    assert ak.to_list(b) == [[None, 1, 2], [], [None, 1], [None], [None, 1, 2, 3]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(6).tolist()  # kernels expect nonnegative entries to be arange
    )

    a = v1_to_v2(
        ak.from_iter(
            [[False, True, None], [], [False, None], [None], [False, True, True, None]],
            highlevel=False,
        )
    )
    b = ak._v2._slicing.prepare_tuple_bool_to_int(a)
    assert ak.to_list(b) == [[1, None], [], [None], [None], [1, 2, None]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(3).tolist()  # kernels expect nonnegative entries to be arange
    )

    a = v1_to_v2(
        ak.from_iter(
            [[None, True, False], [], [None, False], [None], [None, True, True, False]],
            highlevel=False,
        )
    )
    b = ak._v2._slicing.prepare_tuple_bool_to_int(a)
    assert ak.to_list(b) == [[None, 1], [], [None], [None], [None, 1, 2]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(3).tolist()  # kernels expect nonnegative entries to be arange
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import re
import json

import awkward as ak
import numpy as np


def v1v2_id_equal(v1, v2):
    if v1 is None and v2 is None:
        return True
    else:
        raise NotImplementedError("Identities/Identifier equality")


def v1v2_equal(v1, v2):
    assert isinstance(v1, ak.layout.Content) and isinstance(v2, ak._v2.contents.Content)

    if len(v1) != len(v2):
        return False

    if not v1v2_id_equal(v1.identities, v2.identifier):
        return False

    if v1.parameters != v2.parameters:
        return False

    if isinstance(v1, ak.layout.EmptyArray) and isinstance(
        v2, ak._v2.contents.EmptyArray
    ):
        return True

    elif isinstance(v1, ak.layout.NumpyArray) and isinstance(
        v2, ak._v2.contents.NumpyArray
    ):
        v1_data = np.asarray(v1)
        return v1_data.dtype == v2.dtype and np.array_equal(v1_data, v2.data)

    elif isinstance(v1, ak.layout.RegularArray) and isinstance(
        v2, ak._v2.contents.RegularArray
    ):
        return v1v2_equal(v1.content, v2.content) and v1.size == v2.size

    elif isinstance(
        v1, (ak.layout.ListArray32, ak.layout.ListArrayU32, ak.layout.ListArray64)
    ) and isinstance(v2, ak._v2.contents.ListArray):
        v1_starts = np.asarray(v1.starts)
        v1_stops = np.asarray(v1.stops)
        v2_starts = v2.starts.data
        v2_stops = v2.stops.data
        return (
            v1v2_equal(v1.content, v2.content)
            and np.array_equal(v1_starts, v2_starts)
            and v1_starts.dtype == v2_starts.dtype
            and np.array_equal(v1_stops, v2_stops)
            and v1_stops.dtype == v2_stops.dtype
        )

    elif (
        isinstance(
            v1,
            (
                ak.layout.ListOffsetArray32,
                ak.layout.ListOffsetArrayU32,
                ak.layout.ListOffsetArray64,
            ),
        )
        and isinstance(v2, ak._v2.contents.ListOffsetArray)
    ):
        v1_offsets = np.asarray(v1.offsets)
        v2_offsets = v2.offsets.data
        return (
            v1v2_equal(v1.content, v2.content)
            and np.array_equal(v1_offsets, v2_offsets)
            and v1_offsets.dtype == v2_offsets.dtype
        )

    elif isinstance(v1, ak.layout.RecordArray) and isinstance(
        v2, ak._v2.contents.RecordArray
    ):
        return v1.istuple == v2.is_tuple and all(
            v1v2_equal(v1.field(i)[: len(v1)], v2.content(i))
            for i in range(v1.numfields)
        )

    elif (
        isinstance(
            v1,
            (
                ak.layout.IndexedArray32,
                ak.layout.IndexedArrayU32,
                ak.layout.IndexedArray64,
            ),
        )
        and isinstance(v2, ak._v2.contents.IndexedArray)
    ):
        v1_index = np.asarray(v1.index)
        v2_index = v2.index.data
        return (
            v1v2_equal(v1.content, v2.content)
            and np.array_equal(v1_index, v2_index)
            and v1_index.dtype == v2_index.dtype
        )

    elif isinstance(
        v1, (ak.layout.IndexedOptionArray32, ak.layout.IndexedOptionArray64)
    ) and isinstance(v2, ak._v2.contents.IndexedOptionArray):
        v1_index = np.asarray(v1.index)
        v2_index = v2.index.data
        return (
            v1v2_equal(v1.content, v2.content)
            and np.array_equal(v1_index, v2_index)
            and v1_index.dtype == v2_index.dtype
        )

    elif isinstance(v1, ak.layout.ByteMaskedArray) and isinstance(
        v2, ak._v2.contents.ByteMaskedArray
    ):
        v1_mask = np.asarray(v1.mask)
        v2_mask = v2.mask.data
        return (
            v1v2_equal(v1.content, v2.content)
            and np.array_equal(v1_mask, v2_mask)
            and v1_mask.dtype == v2_mask.dtype
            and v1.valid_when == v2.valid_when
        )

    elif isinstance(v1, ak.layout.BitMaskedArray) and isinstance(
        v2, ak._v2.contents.BitMaskedArray
    ):
        v1_mask = np.asarray(v1.mask)
        v2_mask = v2.mask.data
        return (
            v1v2_equal(v1.content, v2.content)
            and np.array_equal(v1_mask, v2_mask)
            and v1_mask.dtype == v2_mask.dtype
            and v1.valid_when == v2.valid_when
            and v1.lsb_order == v2.lsb_order
        )

    elif isinstance(v1, ak.layout.UnmaskedArray) and isinstance(
        v2, ak._v2.contents.UnmaskedArray
    ):
        return v1v2_equal(v1.content, v2.content)

    elif (
        isinstance(
            v1,
            (
                ak.layout.UnionArray8_32,
                ak.layout.UnionArray8_U32,
                ak.layout.UnionArray8_64,
            ),
        )
        and isinstance(v2, ak._v2.contents.UnionArray)
    ):
        v1_tags = np.asarray(v1.tags)
        v2_tags = v2.tags.data
        v1_index = np.asarray(v1.index)
        v2_index = v2.index.data
        return (
            np.array_equal(v1_tags, v2_tags)
            and v1_tags.dtype == v2_tags.dtype
            and np.array_equal(v1_index, v2_index)
            and v1_index.dtype == v2_index.dtype
            and all(
                v1v2_equal(v1.content(i), v2.content(i)) for i in range(v1.numcontents)
            )
        )

    elif isinstance(v1, ak.layout.VirtualArray) and isinstance(
        v2, ak._v2.contents.VirtualArray
    ):
        raise NotImplementedError("VirtualArray")

    else:
        raise AssertionError("{0} vs {1}".format(type(v1), type(v2)))


def v1_to_v2_id(v1):
    if v1 is None:
        return None
    else:
        raise NotImplementedError("Identities to Identifier")


def fix(array):
    if issubclass(array.dtype.type, np.signedinteger):
        if array.dtype.itemsize == 8:
            return array.view(np.int64)
        elif array.dtype.itemsize == 4:
            return array.view(np.int32)
        elif array.dtype.itemsize == 2:
            return array.view(np.int16)
        elif array.dtype.itemsize == 1:
            return array.view(np.int8)
    elif issubclass(array.dtype.type, np.unsignedinteger):
        if array.dtype.itemsize == 8:
            return array.view(np.uint64)
        elif array.dtype.itemsize == 4:
            return array.view(np.uint32)
        elif array.dtype.itemsize == 2:
            return array.view(np.uint16)
        elif array.dtype.itemsize == 1:
            return array.view(np.uint8)
    else:
        return array


def v1_to_v2(v1):
    assert isinstance(v1, ak.layout.Content)

    if isinstance(v1, ak.layout.EmptyArray):
        return ak._v2.contents.EmptyArray(
            identifier=v1_to_v2_id(v1.identities), parameters=v1.parameters
        )

    elif isinstance(v1, ak.layout.NumpyArray):
        primitive = v1.form.primitive
        if primitive == "datetime64" or primitive == "timedelta64":
            return ak._v2.contents.NumpyArray(
                fix(np.asarray(v1.view_int64)).view(np.dtype(v1.format)),
                identifier=v1_to_v2_id(v1.identities),
                parameters=v1.parameters,
            )
        else:
            return ak._v2.contents.NumpyArray(
                fix(np.asarray(v1)),
                identifier=v1_to_v2_id(v1.identities),
                parameters=v1.parameters,
            )

    elif isinstance(v1, ak.layout.RegularArray):
        return ak._v2.contents.RegularArray(
            v1_to_v2(v1.content),
            v1.size,
            len(v1),
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(
        v1, (ak.layout.ListArray32, ak.layout.ListArrayU32, ak.layout.ListArray64)
    ):
        return ak._v2.contents.ListArray(
            ak._v2.index.Index(fix(np.asarray(v1.starts))),
            ak._v2.index.Index(fix(np.asarray(v1.stops))),
            v1_to_v2(v1.content),
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(
        v1,
        (
            ak.layout.ListOffsetArray32,
            ak.layout.ListOffsetArrayU32,
            ak.layout.ListOffsetArray64,
        ),
    ):
        return ak._v2.contents.ListOffsetArray(
            ak._v2.index.Index(fix(np.asarray(v1.offsets))),
            v1_to_v2(v1.content),
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(v1, ak.layout.RecordArray):
        return ak._v2.contents.RecordArray(
            [v1_to_v2(x) for x in v1.contents],
            v1.recordlookup,
            len(v1),
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(
        v1,
        (
            ak.layout.IndexedArray32,
            ak.layout.IndexedArrayU32,
            ak.layout.IndexedArray64,
        ),
    ):
        return ak._v2.contents.IndexedArray(
            ak._v2.index.Index(fix(np.asarray(v1.index))),
            v1_to_v2(v1.content),
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(
        v1, (ak.layout.IndexedOptionArray32, ak.layout.IndexedOptionArray64)
    ):
        return ak._v2.contents.IndexedOptionArray(
            ak._v2.index.Index(fix(np.asarray(v1.index))),
            v1_to_v2(v1.content),
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(v1, ak.layout.ByteMaskedArray):
        return ak._v2.contents.ByteMaskedArray(
            ak._v2.index.Index(fix(np.asarray(v1.mask))),
            v1_to_v2(v1.content),
            v1.valid_when,
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(v1, ak.layout.BitMaskedArray):
        return ak._v2.contents.BitMaskedArray(
            ak._v2.index.Index(fix(np.asarray(v1.mask))),
            v1_to_v2(v1.content),
            v1.valid_when,
            len(v1),
            v1.lsb_order,
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(v1, ak.layout.UnmaskedArray):
        return ak._v2.contents.UnmaskedArray(
            v1_to_v2(v1.content),
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(
        v1,
        (
            ak.layout.UnionArray8_32,
            ak.layout.UnionArray8_U32,
            ak.layout.UnionArray8_64,
        ),
    ):
        return ak._v2.contents.UnionArray(
            ak._v2.index.Index(fix(np.asarray(v1.tags))),
            ak._v2.index.Index(fix(np.asarray(v1.index))),
            [v1_to_v2(x) for x in v1.contents],
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    elif isinstance(v1, ak.layout.VirtualArray):
        form = v1.generator.form
        if form is not None:
            form = ak._v2.forms.from_json(form.tojson())

        cache = v1.cache
        if cache is not None:
            cache = cache.mutablemapping

        if isinstance(v1.generator, ak.layout.SliceGenerator):
            # cheesy way to get the slice
            repr_gen = repr(v1.generator)
            range_slice = re.search(
                r"<slice>\[(-?[0-9]+):(-?[0-9]+)\]</slice>", repr_gen
            )
            field_slice = re.search(r"<slice>\[(.*)\]</slice>", repr_gen)
            if range_slice is not None:
                slc = slice(int(range_slice.group(1)), int(range_slice.group(2)))
            else:
                slc = eval(field_slice.group(1))

            generator = ak._v2.contents.virtualarray.SliceGenerator(
                v1.generator.length,
                form,
                v1_to_v2(v1.generator.content),
                slc,
            )

        else:
            generator = ak._v2.contents.virtualarray.FunctionGenerator(
                v1.generator.length,
                form,
                v1.generator.callable,
                v1.generator.args,
                v1.generator.kwargs,
            )

        return ak._v2.contents.VirtualArray(
            generator,
            cache=cache,
            cache_key=v1.cache_key,
            identifier=v1_to_v2_id(v1.identities),
            parameters=v1.parameters,
        )

    else:
        raise AssertionError(type(v1))


def v2_to_v1_id(v1):
    if v1 is None:
        return None
    else:
        raise NotImplementedError("Identifier to Identities")


def v2_to_v1(v2):
    assert isinstance(v2, ak._v2.contents.Content)

    if isinstance(v2, ak._v2.contents.EmptyArray):
        return ak.layout.EmptyArray(
            identities=v2_to_v1_id(v2.identifier), parameters=v2.parameters
        )

    elif isinstance(v2, ak._v2.contents.NumpyArray):
        return ak.layout.NumpyArray(
            v2.data, identities=v2_to_v1_id(v2.identifier), parameters=v2.parameters
        )

    elif isinstance(v2, ak._v2.contents.RegularArray):
        return ak.layout.RegularArray(
            v2_to_v1(v2.content),
            v2.size,
            len(v2),
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.ListArray):
        assert v2.starts.dtype == v2.stops.dtype
        if v2.starts.dtype == np.dtype(np.int32):
            ind = ak.layout.Index32
            cls = ak.layout.ListArray32
        elif v2.starts.dtype == np.dtype(np.uint32):
            ind = ak.layout.IndexU32
            cls = ak.layout.ListArrayU32
        elif v2.starts.dtype == np.dtype(np.int64):
            ind = ak.layout.Index64
            cls = ak.layout.ListArray64
        return cls(
            ind(v2.starts.data),
            ind(v2.stops.data),
            v2_to_v1(v2.content),
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.ListOffsetArray):
        if v2.offsets.dtype == np.dtype(np.int32):
            ind = ak.layout.Index32
            cls = ak.layout.ListOffsetArray32
        elif v2.offsets.dtype == np.dtype(np.uint32):
            ind = ak.layout.IndexU32
            cls = ak.layout.ListOffsetArrayU32
        elif v2.offsets.dtype == np.dtype(np.int64):
            ind = ak.layout.Index64
            cls = ak.layout.ListOffsetArray64
        return cls(
            ind(v2.offsets.data),
            v2_to_v1(v2.content),
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.RecordArray):
        return ak.layout.RecordArray(
            [v2_to_v1(x) for x in v2.contents],
            None if v2.is_tuple else v2.keys,
            len(v2),
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.IndexedArray):
        if v2.index.dtype == np.dtype(np.int32):
            ind = ak.layout.Index32
            cls = ak.layout.IndexedArray32
        elif v2.index.dtype == np.dtype(np.uint32):
            ind = ak.layout.IndexU32
            cls = ak.layout.IndexedArrayU32
        elif v2.index.dtype == np.dtype(np.int64):
            ind = ak.layout.Index64
            cls = ak.layout.IndexedArray64
        return cls(
            ind(v2.index.data),
            v2_to_v1(v2.content),
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.IndexedOptionArray):
        if v2.index.dtype == np.dtype(np.int32):
            ind = ak.layout.Index32
            cls = ak.layout.IndexedOptionArray32
        elif v2.index.dtype == np.dtype(np.uint32):
            ind = ak.layout.IndexU32
            cls = ak.layout.IndexedOptionArrayU32
        elif v2.index.dtype == np.dtype(np.int64):
            ind = ak.layout.Index64
            cls = ak.layout.IndexedOptionArray64
        return cls(
            ind(v2.index.data),
            v2_to_v1(v2.content),
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.ByteMaskedArray):
        return ak.layout.ByteMaskedArray(
            ak.layout.Index8(v2.mask.data),
            v2_to_v1(v2.content),
            v2.valid_when,
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.BitMaskedArray):
        return ak.layout.BitMaskedArray(
            ak.layout.IndexU8(v2.mask.data),
            v2_to_v1(v2.content),
            v2.valid_when,
            len(v2),
            v2.lsb_order,
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.UnmaskedArray):
        return ak.layout.UnmaskedArray(
            v2_to_v1(v2.content),
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.UnionArray):
        if v2.index.dtype == np.dtype(np.int32):
            ind = ak.layout.Index32
            cls = ak.layout.UnionArray8_32
        elif v2.index.dtype == np.dtype(np.uint32):
            ind = ak.layout.IndexU32
            cls = ak.layout.UnionArray8_U32
        elif v2.index.dtype == np.dtype(np.int64):
            ind = ak.layout.Index64
            cls = ak.layout.UnionArray8_64
        return cls(
            ak.layout.Index8(v2.tags.data),
            ind(v2.index.data),
            [v2_to_v1(x) for x in v2.contents],
            identities=v2_to_v1_id(v2.identifier),
            parameters=v2.parameters,
        )

    elif isinstance(v2, ak._v2.contents.VirtualArray):
        if v2.generator.form is None:
            form = None
        else:
            form = ak.forms.Form.fromjson(json.dumps(v2.generator.form.tolist()))

        def function(v2generator):
            array = v2generator.generate()
            if isinstance(array, (ak._v2.contents.Content, ak._v2.record.Record)):
                return v2_to_v1(array)
            else:
                return array

        generator = ak.layout.ArrayGenerator(
            function, (v2.generator,), {}, form=form, length=v2.generator.length
        )

        return ak.layout.VirtualArray(
            generator, None, cache_key=v2.cache_key, parameters=v2.parameters
        )

    else:
        raise AssertionError(type(v2))


def v1_to_v2_index(v1):
    assert isinstance(
        v1,
        (
            ak.layout.IndexU8,
            ak.layout.IndexU32,
            ak.layout.Index8,
            ak.layout.Index32,
            ak.layout.Index64,
        ),
    )

    if isinstance(v1, ak.layout.IndexU8):
        return ak._v2.index.IndexU8(np.asarray(v1))

    elif isinstance(v1, ak.layout.IndexU32):
        return ak._v2.index.IndexU32(np.asarray(v1))

    elif isinstance(v1, ak.layout.Index8):
        return ak._v2.index.Index8(np.asarray(v1))

    elif isinstance(v1, ak.layout.Index32):
        return ak._v2.index.Index32(np.asarray(v1))

    elif isinstance(v1, ak.layout.Index64):
        return ak._v2.index.Index64(np.asarray(v1))

    else:
        raise AssertionError(type(v1))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import itertools

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

content = ak.layout.NumpyArray(
    np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
)
offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
regulararray = ak.layout.RegularArray(listoffsetarray, 2, zeros_length=0)
starts = ak.layout.Index64(np.array([0, 1]))
stops = ak.layout.Index64(np.array([2, 3]))
listarray = ak.layout.ListArray64(starts, stops, regulararray)


def test_simple_type():
    assert str(ak.type(content)) == "float64"


def test_type():
    assert str(ak.type(regulararray)) == "2 * var * float64"


def test_iteration():
    assert ak.to_list(regulararray) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]


def test_tojson():
    assert (
        ak.to_json(regulararray)
        == "[[[0.0,1.1,2.2],[]],[[3.3,4.4],[5.5]],[[6.6,7.7,8.8,9.9],[]]]"
    )


def test_getitem_at():
    assert ak.to_list(regulararray[0]) == [[0.0, 1.1, 2.2], []]
    assert ak.to_list(regulararray[1]) == [[3.3, 4.4], [5.5]]
    assert ak.to_list(regulararray[2]) == [[6.6, 7.7, 8.8, 9.9], []]


def test_getitem_range():
    assert ak.to_list(regulararray[1:]) == [
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert ak.to_list(regulararray[:-1]) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]]


def test_getitem():
    assert ak.to_list(regulararray[(0,)]) == [[0.0, 1.1, 2.2], []]
    assert ak.to_list(regulararray[(1,)]) == [[3.3, 4.4], [5.5]]
    assert ak.to_list(regulararray[(2,)]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert ak.to_list(regulararray[(slice(1, None, None),)]) == [
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]
    assert ak.to_list(regulararray[(slice(None, -1, None),)]) == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
    ]


def test_getitem_deeper():
    assert ak.to_list(listarray) == [
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]],
        [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]],
    ]

    assert ak.to_list(listarray[0, 0, 0]) == [0.0, 1.1, 2.2]
    assert ak.to_list(listarray[0, 0, 1]) == []
    assert ak.to_list(listarray[0, 1, 0]) == [3.3, 4.4]
    assert ak.to_list(listarray[0, 1, 1]) == [5.5]
    assert ak.to_list(listarray[1, 0, 0]) == [3.3, 4.4]
    assert ak.to_list(listarray[1, 0, 1]) == [5.5]
    assert ak.to_list(listarray[1, 1, 0]) == [6.6, 7.7, 8.8, 9.9]
    assert ak.to_list(listarray[1, 1, 1]) == []

    assert ak.to_list(listarray[0, 0, 0:]) == [[0.0, 1.1, 2.2], []]
    assert ak.to_list(listarray[0, 0, 1:]) == [[]]
    assert ak.to_list(listarray[0, 1, 0:]) == [[3.3, 4.4], [5.5]]
    assert ak.to_list(listarray[0, 1, 1:]) == [[5.5]]
    assert ak.to_list(listarray[1, 0, 0:]) == [[3.3, 4.4], [5.5]]
    assert ak.to_list(listarray[1, 0, 1:]) == [[5.5]]
    assert ak.to_list(listarray[1, 1, 0:]) == [[6.6, 7.7, 8.8, 9.9], []]
    assert ak.to_list(listarray[1, 1, 1:]) == [[]]

    assert ak.to_list(listarray[[1], 0, 0:]) == [[[3.3, 4.4], [5.5]]]
    assert ak.to_list(listarray[[1, 0], 0, 0:]) == [
        [[3.3, 4.4], [5.5]],
        [[0.0, 1.1, 2.2], []],
    ]

    assert ak.to_list(listarray[:, :, [0, 1]]) == [
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]],
        [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]],
    ]
    assert ak.to_list(listarray[:, :, [1, 0]]) == [
        [[[], [0.0, 1.1, 2.2]], [[5.5], [3.3, 4.4]]],
        [[[5.5], [3.3, 4.4]], [[], [6.6, 7.7, 8.8, 9.9]]],
    ]
    assert ak.to_list(listarray[:, :, [1, 0, 1]]) == [
        [[[], [0.0, 1.1, 2.2], []], [[5.5], [3.3, 4.4], [5.5]]],
        [[[5.5], [3.3, 4.4], [5.5]], [[], [6.6, 7.7, 8.8, 9.9], []]],
    ]
    assert ak.to_list(listarray[:, :2, [0, 1]]) == [
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]],
        [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]],
    ]

    assert ak.to_list(listarray[:1, [0, 0, 1, 1], [0, 1, 0, 1]]) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5]]
    ]
    assert ak.to_list(listarray[:1, [1, 1, 0, 0], [1, 0, 1, 0]]) == [
        [[5.5], [3.3, 4.4], [], [0.0, 1.1, 2.2]]
    ]


content2 = ak.layout.NumpyArray(np.arange(2 * 3 * 5 * 7).reshape(-1, 7))
regulararrayA = ak.layout.RegularArray(content2, 5, zeros_length=0)
regulararrayB = ak.layout.RegularArray(regulararrayA, 3, zeros_length=0)
modelA = np.arange(2 * 3 * 5 * 7).reshape(2 * 3, 5, 7)
modelB = np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7)


def test_numpy():
    assert ak.to_list(regulararrayA) == ak.to_list(modelA)
    assert ak.to_list(regulararrayB) == ak.to_list(modelB)

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((0, 1, 4, -5), depth):
            assert ak.to_list(modelA[cuts]) == ak.to_list(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            (slice(None), slice(1, None), slice(None, -1), slice(None, None, 2)), depth
        ):
            assert ak.to_list(modelA[cuts]) == ak.to_list(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            (slice(1, None), slice(None, -1), 2, -2), depth
        ):
            assert ak.to_list(modelA[cuts]) == ak.to_list(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            ([2, 0, 0, 1], [1, -2, 0, -1], 2, -2), depth
        ):
            assert ak.to_list(modelA[cuts]) == ak.to_list(regulararrayA[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(
            ([2, 0, 0, 1], [1, -2, 0, -1], slice(1, None), slice(None, -1)), depth
        ):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert ak.to_list(modelA[cuts]) == ak.to_list(regulararrayA[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-2, -1, 0, 1, 1), depth):
            assert ak.to_list(modelB[cuts]) == ak.to_list(regulararrayB[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations(
            (-1, 0, 1, slice(1, None), slice(None, -1)), depth
        ):
            assert ak.to_list(modelB[cuts]) == ak.to_list(regulararrayB[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations(
            (-1, 0, [1, 0, 0, 1], [0, 1, -1, 1], slice(None, -1)), depth
        ):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert ak.to_list(modelB[cuts]) == ak.to_list(regulararrayB[cuts])


def test_setidentities():
    regulararray.setidentities()
    assert np.asarray(regulararray.identities).tolist() == [[0], [1], [2]]
    assert np.asarray(regulararray.content.identities).tolist() == [
        [0, 0],
        [0, 1],
        [1, 0],
        [1, 1],
        [2, 0],
        [2, 1],
    ]
    assert np.asarray(regulararray.content.content.identities).tolist() == [
        [0, 0, 0],  # 0.0
        [0, 0, 1],  # 1.1
        [0, 0, 2],  # 2.2
        # [0, 1,  ],   # (empty list)
        [1, 0, 0],  # 3.3
        [1, 0, 1],  # 4.4
        [1, 1, 0],  # 5.5
        [2, 0, 0],  # 6.6
        [2, 0, 1],  # 7.7
        [2, 0, 2],  # 8.8
        [2, 0, 3],
    ]  # 9.9
    # [2, 1,  ],   # (empty list)

    regulararrayB.setidentities()
    assert np.asarray(regulararrayB.identities).tolist() == [[0], [1]]
    assert np.asarray(regulararrayB.content.identities).tolist() == [
        [0, 0],
        [0, 1],
        [0, 2],
        [1, 0],
        [1, 1],
        [1, 2],
    ]
